#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2021 Roberto Chamorro / project contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Special thanks to
# https://github.com/bear/python-twitter/
# and
# https://github.com/halcy/Mastodon.py

import os
import sys
import time
import yaml
import shutil
import logging
import argparse

from requests_oauthlib import OAuth1

from .i18n import _
from . import logger
from .__init__ import __version__


class User(object):
    from ._twitter import get_tweets
    from ._twitter import _get_tweets
    from ._twitter import _get_tweets_v2
    from ._twitter import _get_twitter_info

    from ._pin import pin_pleroma
    from ._pin import unpin_pleroma
    from ._pin import get_pinned_tweet
    from ._pin import _get_pinned_tweet_id

    from ._pleroma import post_pleroma
    from ._pleroma import update_pleroma
    from ._pleroma import get_date_last_pleroma_post

    from ._utils import force_date
    from ._utils import guess_type
    from ._utils import check_pinned
    from ._utils import random_string
    from ._utils import _get_instance_info
    from ._utils import replace_vars_in_str

    from ._processing import process_tweets

    from ._processing import _expand_urls
    from ._processing import _get_media_url
    from ._processing import _process_polls
    from ._processing import _replace_nitter
    from ._processing import _download_media
    from ._processing import _replace_mentions
    from ._processing import _get_best_bitrate_video

    def __init__(self, user_cfg: dict, cfg: dict, base_path: str):
        self.posts = None
        self.tweets = None
        self.first_time = False
        self.display_name = None
        self.last_post_pleroma = None
        self.profile_image_url = None
        self.profile_banner_url = None
        valid_visibility = ("public", "unlisted", "private", "direct")
        default_cfg_attributes = {
            "twitter_base_url": "https://api.twitter.com/1.1",
            "twitter_base_url_v2": "https://api.twitter.com/2",
            "nitter_base_url": "https://nitter.net",
            "pleroma_base_url": None,
            "nitter": False,
            "twitter_token": None,
            "signature": False,
            "media_upload": False,
            "sensitive": False,
            "max_tweets": 50,
            "delay_post": 0.5,
            "visibility": "unlisted",
            "hashtags": [],
            "tweet_ids": [],
            "rich_text": False,
            "twitter_bio": True,
            "include_rts": True,
            "include_replies": True,
            "consumer_key": None,
            "consumer_secret": None,
            "access_token_key": None,
            "access_token_secret": None,
            "original_date": False,
            "original_date_format": "%Y-%m-%d %H:%M",
        }
        # iterate attrs defined in config
        for attribute in default_cfg_attributes:
            if attribute in cfg:
                self.__setattr__(attribute, cfg[attribute])
            if not hasattr(self, attribute):
                self.__setattr__(attribute, default_cfg_attributes[attribute])
        for user_attribute in user_cfg:
            self.__setattr__(user_attribute, user_cfg[user_attribute])

        twitter_url = (
            self.nitter_base_url if self.nitter else "http://twitter.com"
        )
        self.twitter_url = f"{twitter_url}/{self.twitter_username}"
        if self.rich_text:
            self.content_type = "text/markdown"
        if not self.pleroma_base_url:
            raise KeyError(
                _("No Pleroma URL defined in config! [pleroma_base_url]")
            )
        if self.visibility not in valid_visibility:
            raise KeyError(
                _("Visibility not supported! Values allowed are: {}").format(
                    ", ".join(valid_visibility)
                )
            )
        try:
            self.fields = self.replace_vars_in_str(str(user_cfg["fields"]))
            self.fields = eval(self.fields)
        except KeyError:
            self.fields = []
        self.bio_text = self.replace_vars_in_str(str(user_cfg["bio_text"]))
        # Auth
        self.header_pleroma = {"Authorization": f"Bearer {self.pleroma_token}"}
        self.header_twitter = {"Authorization": f"Bearer {self.twitter_token}"}

        if all(
            [
                self.consumer_key,
                self.consumer_secret,
                self.access_token_key,
                self.access_token_secret,
            ]
        ):
            self.auth = OAuth1(
                self.consumer_key,
                self.consumer_secret,
                self.access_token_key,
                self.access_token_secret,
            )
        else:
            self.auth = None
            logger.debug(
                _(
                    "Some or all OAuth 1.0a tokens missing, "
                    "falling back to application-only authentication"
                )
            )

        self.pinned_tweet_id = self._get_pinned_tweet_id()

        # Filesystem
        self.base_path = base_path
        self.users_path = os.path.join(self.base_path, "users")
        self.users_path = os.path.join(self.base_path, "users")
        self.user_path = os.path.join(self.users_path, self.twitter_username)
        self.tweets_temp_path = os.path.join(self.user_path, "tweets")
        self.avatar_path = os.path.join(self.user_path, "profile.jpg")
        self.header_path = os.path.join(self.user_path, "banner.jpg")
        os.makedirs(self.users_path, exist_ok=True)
        os.makedirs(self.user_path, exist_ok=True)
        os.makedirs(self.tweets_temp_path, exist_ok=True)
        # Get Twitter info on instance creation
        self._get_twitter_info()
        self._get_instance_info()

        return


def get_args(sysargs):
    """
    Supports the command-line arguments listed below.
    """
    parser = argparse.ArgumentParser(
        description=(
            _(
                "Bot for mirroring one or multiple Twitter accounts "
                "in Pleroma/Mastodon."
            )
        )
    )

    parser.add_argument(
        "-c",
        "--config",
        required=False,
        action="store",
        help=(
            _(
                "path of config file (config.yml) to use and parse. If not"
                " specified, it will try to find it in the current working "
                "directory."
            )
        ),
    )

    parser.add_argument(
        "-l",
        "--log",
        required=False,
        action="store",
        help=(
            _(
                "path of log file (error.log) to create. If not"
                " specified, it will try to store it at your config file path"
            )
        ),
    )

    parser.add_argument(
        "-n",
        "--noProfile",
        required=False,
        action="store_true",
        help=(
            _(
                "skips Fediverse profile update (no background "
                "image, profile image, bio text, etc.)"
            )
        ),
    )

    parser.add_argument(
        "--forceDate",
        nargs="?",
        required=False,
        action="store",
        const="all",
        help=(
            _(
                "forces the tweet retrieval to start from a "
                "specific date. The twitter_username value "
                "(FORCEDATE) can be supplied to only force it for "
                "that particular user in the config"
            )
        ),
    )

    parser.add_argument(
        "-s",
        "--skipChecks",
        required=False,
        action="store_true",
        help=(_("skips first run checks")),
    )

    parser.add_argument("--verbose", "-v", action="count", default=0)

    parser.add_argument(
        "--version", action="version", version=f"{__version__}"
    )

    args, extra = parser.parse_known_args(sysargs)
    return args


def main():
    # Convert legacy flag to proper flag format
    mangle_args = "noProfile"
    arguments = [
        "--" + arg if arg in mangle_args else arg for arg in sys.argv[1:]
    ]
    args = get_args(sysargs=arguments)

    if args.verbose > 0:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(_("Debug logging enabled"))

    try:
        base_path = os.getcwd()
        if args.config:
            config_path = args.config
            base_path, cfg_file = os.path.split(os.path.abspath(config_path))
        else:
            config_path = os.path.join(base_path, "config.yml")

        with open(config_path, "r") as stream:
            config = yaml.safe_load(stream)
        user_dict = config["users"]
        users_path = os.path.join(base_path, "users")
        # TODO: Merge tweets of multiple accounts and order them by date
        for user_item in user_dict[:]:
            user_item["skip_pin"] = False
            if isinstance(user_item["twitter_username"], list):
                warn_msg = _(
                    "Multiple twitter users for one Fediverse account, "
                    "skipping profile and pinned tweet."
                )
                logger.warning(warn_msg)
                user_item["skip_pin"] = True
                for twitter_user in user_item["twitter_username"]:
                    new_user = dict(user_item)
                    new_user["twitter_username"] = twitter_user
                    user_dict.append(new_user)
                user_dict.remove(user_item)

        for user_item in user_dict:
            first_time = False
            logger.info("======================================")
            logger.info(
                _("Processing user:\t{}").format(user_item["pleroma_username"])
            )
            user_path = os.path.join(users_path, user_item["twitter_username"])

            if not os.path.exists(user_path):
                first_time_msg = _(
                    "It seems like pleroma-bot is running for the "
                    "first time for this user"
                )
                logger.info(first_time_msg)
                first_time = True
            user = User(user_item, config, base_path)
            if first_time and not args.skipChecks:
                user.first_time = True
            if (
                (args.forceDate and args.forceDate == user.twitter_username)
                or args.forceDate == "all"
                or user.first_time
            ) and not args.skipChecks:
                date_pleroma = user.force_date()
            else:
                date_pleroma = user.get_date_last_pleroma_post()

            if user.tweet_ids:
                tweets = {
                    "data": [],
                    "includes": {},
                    "meta": {"result_count": len(user.tweet_ids)},
                }
                includes = ["users", "tweets", "media", "polls"]
                for include in includes:
                    try:
                        _include = tweets["includes"][include]
                    except KeyError:
                        tweets["includes"].update({include: []})
                for tweet_id in user.tweet_ids:
                    next_tweet = user._get_tweets("v2", tweet_id=tweet_id)
                    includes = ["users", "tweets", "media", "polls"]
                    for include in includes:
                        try:
                            _include = next_tweet["includes"][include]
                            _include = _include
                        except KeyError:
                            next_tweet["includes"].update({include: []})
                    tweets["data"].append(next_tweet["data"])
                    for user_tweet in next_tweet["includes"]["users"]:
                        tweets["includes"]["users"].append(user_tweet)
                    for tweet_include in next_tweet["includes"]["tweets"]:
                        tweets["includes"]["tweets"].append(tweet_include)
                    for media in next_tweet["includes"]["media"]:
                        tweets["includes"]["media"].append(media)
                    for poll in next_tweet["includes"]["polls"]:
                        tweets["includes"]["polls"].append(poll)
            else:
                tweets = user.get_tweets(start_time=date_pleroma)
            logger.debug(f"tweets: \t {tweets}")

            if "meta" not in tweets:
                error_msg = _(
                    "Unable to retrieve tweets. Is the account protected?"
                    " If so, you need to provide the following OAuth 1.0a"
                    " fields in the user config:\n - consumer_key \n "
                    "- consumer_secret \n - access_token_key \n "
                    "- access_token_secret"
                )
                logger.error(error_msg)

            if tweets["meta"]["result_count"] > 0:
                logger.info(
                    _("tweet count: \t {}").format(len(tweets["data"]))
                )
                # Put oldest first to iterate them and post them in order
                tweets["data"].reverse()
                tweets_to_post = user.process_tweets(tweets)
                logger.debug(f"tweets_processed: \t {tweets_to_post['data']}")
                tweet_counter = 0
                for tweet in tweets_to_post["data"]:
                    tweet_counter += 1
                    logger.info(
                        f"({tweet_counter}/{len(tweets_to_post['data'])})"
                    )
                    user.post_pleroma(
                        (tweet["id"], tweet["text"], tweet["created_at"]),
                        tweet["polls"],
                        tweet["possibly_sensitive"],
                    )

                    time.sleep(user.delay_post)
            if not user.skip_pin:
                user.check_pinned()

            if not args.noProfile:
                if user.skip_pin:
                    logger.warning(
                        _("Multiple twitter users, not updating profile")
                    )
                else:
                    user.update_pleroma()
            # Clean-up
            shutil.rmtree(user.tweets_temp_path)
    except Exception:
        logger.error(_("Exception occurred"), exc_info=True)
        return 1

    return 0


def init():
    # Convert legacy flag to proper flag format
    mangle_args = "noProfile"
    arguments = [
        "--" + arg if arg in mangle_args else arg for arg in sys.argv[1:]
    ]
    args = get_args(sysargs=arguments)
    if args.log:
        log_path = args.log
    elif args.config:
        base_path, cfg_file = os.path.split(os.path.abspath(args.config))
        log_path = os.path.join(base_path, "error.log")
    else:
        log_path = os.path.join(os.getcwd(), "error.log")
    f_handler = logging.FileHandler(log_path)
    f_handler.setLevel(logging.ERROR)
    f_format = logging.Formatter(
        "%(asctime)s %(name)s %(levelname)s: %(message)s"
    )
    f_handler.setFormatter(f_format)
    logger.addHandler(f_handler)
    if __name__ == "__main__":
        sys.exit(main())


init()



import os
import gettext

localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "locale")
translate = gettext.translation("pleroma_bot", localedir, fallback=True)
_ = translate.gettext



import json
import requests


from pleroma_bot.i18n import _
from pleroma_bot._utils import spinner


def _get_twitter_info(self):
    """Updates User object attributes with current Twitter info

    This includes:

    * Bio text
    * Profile image url
    * Banner image url
    * Screen name

    :return: None
    """
    twitter_user_url = (
        f"{self.twitter_base_url}"
        f"/users/show.json?screen_name="
        f"{self.twitter_username}"
    )
    response = requests.get(
        twitter_user_url, headers=self.header_twitter, auth=self.auth
    )
    if not response.ok:
        response.raise_for_status()
    user_twitter = json.loads(response.text)
    self.bio_text = (
        f'{self.bio_text}{user_twitter["description"]}'
        if self.twitter_bio
        else f"{self.bio_text}"
    )
    # Check if user has profile image
    if "profile_image_url_https" in user_twitter.keys():
        self.profile_image_url = user_twitter["profile_image_url_https"]
    # Check if user has banner image
    if "profile_banner_url" in user_twitter.keys():
        base_banner_url = user_twitter["profile_banner_url"]
        self.profile_banner_url = f"{base_banner_url}/1500x500"
    self.display_name = user_twitter["name"]
    return


def _get_tweets(self, version: str, tweet_id=None, start_time=None):
    """Gathers last 'max_tweets' tweets from the user and returns them
    as an dict
    :param version: Twitter API version to use to retrieve the tweets
    :type version: string
    :param tweet_id: Tweet ID to retrieve
    :type tweet_id: int

    :returns: last 'max_tweets' tweets
    :rtype: dict
    """
    if version == "v1.1":
        if tweet_id:
            twitter_status_url = (
                f"{self.twitter_base_url}/statuses/"
                f"show.json?id={str(tweet_id)}"
            )
            response = requests.get(
                twitter_status_url, headers=self.header_twitter, auth=self.auth
            )
            if not response.ok:
                response.raise_for_status()
            tweet = json.loads(response.text)
            return tweet
        else:
            twitter_status_url = (
                f"{self.twitter_base_url}"
                f"/statuses/user_timeline.json?screen_name="
                f"{self.twitter_username}"
                f"&count={str(self.max_tweets)}&include_rts=true"
            )
            response = requests.get(
                twitter_status_url, headers=self.header_twitter, auth=self.auth
            )
            if not response.ok:
                response.raise_for_status()
            tweets = json.loads(response.text)
            return tweets
    elif version == "v2":
        tweets_v2 = self._get_tweets_v2(
            tweet_id=tweet_id, start_time=start_time
        )
        return tweets_v2
    else:
        raise ValueError(_("API version not supported: {}").format(version))


def _get_tweets_v2(
    self,
    start_time,
    tweet_id=None,
    next_token=None,
    previous_token=None,
    tweets_v2=None,
):
    # Tweet number must be between 10 and 100
    if not (100 >= self.max_tweets > 10):
        global _
        error_msg = _(
            "max_tweets must be between 10 and 100. max_tweets: {}"
        ).format(self.max_tweets)
        raise ValueError(error_msg)
    params = {}
    previous_token = next_token
    if tweet_id:
        url = f"{self.twitter_base_url_v2}/tweets/{tweet_id}"
    else:
        url = (
            f"{self.twitter_base_url_v2}/users/by?"
            f"usernames={self.twitter_username}"
        )
        response = requests.get(
            url, headers=self.header_twitter, auth=self.auth
        )
        if not response.ok:
            response.raise_for_status()
        response = json.loads(response.text)
        twitter_id = response["data"][0]["id"]

        url = f"{self.twitter_base_url_v2}/users/{twitter_id}/tweets"
        if next_token:
            params.update({"pagination_token": next_token})

        params.update(
            {
                "start_time": start_time,
                "max_results": self.max_tweets,
            }
        )

    params.update(
        {
            "poll.fields": "duration_minutes,end_datetime,id,options,"
            "voting_status",
            "media.fields": "duration_ms,height,media_key,"
            "preview_image_url,type,url,width,"
            "public_metrics",
            "expansions": "attachments.poll_ids,"
            "attachments.media_keys,author_id,"
            "entities.mentions.username,geo.place_id,"
            "in_reply_to_user_id,referenced_tweets.id,"
            "referenced_tweets.id.author_id",
            "tweet.fields": "attachments,author_id,"
            "context_annotations,conversation_id,"
            "created_at,entities,"
            "geo,id,in_reply_to_user_id,lang,"
            "public_metrics,"
            "possibly_sensitive,referenced_tweets,"
            "source,text,"
            "withheld",
        }
    )

    response = requests.get(
        url, headers=self.header_twitter, params=params, auth=self.auth
    )
    if not response.ok:
        response.raise_for_status()

    if tweets_v2:
        next_tweets = response.json()

        includes = ["users", "tweets", "media", "polls"]
        for include in includes:
            try:
                _ = tweets_v2["includes"][include]
            except KeyError:
                tweets_v2["includes"].update({include: []})
            try:
                _ = next_tweets["includes"][include]
            except KeyError:
                next_tweets["includes"].update({include: []})

        for tweet in next_tweets["data"]:
            tweets_v2["data"].append(tweet)
        for user in next_tweets["includes"]["users"]:
            tweets_v2["includes"]["users"].append(user)
        for tweet_include in next_tweets["includes"]["tweets"]:
            tweets_v2["includes"]["tweets"].append(tweet_include)
        for media in next_tweets["includes"]["media"]:
            tweets_v2["includes"]["media"].append(media)
        for poll in next_tweets["includes"]["polls"]:
            tweets_v2["includes"]["polls"].append(poll)
        tweets_v2["meta"] = next_tweets["meta"]
    else:
        tweets_v2 = json.loads(response.text)

    try:
        next_token = response.json()["meta"]["next_token"]
        if next_token and next_token != previous_token:
            self._get_tweets_v2(
                start_time=start_time,
                tweets_v2=tweets_v2,
                next_token=next_token,
                previous_token=previous_token,
            )
    except KeyError:
        pass

    return tweets_v2


@spinner(_("Gathering tweets... "))
def get_tweets(self, start_time):
    return self._get_tweets("v2", start_time=start_time)



import os
import re
import json
import html
import shutil
import requests
import mimetypes

from ._utils import spinner

# Try to import libmagic
# if it fails just use mimetypes
try:
    import magic
except ImportError:
    magic = None

from . import logger
from .i18n import _


@spinner(_("Processing tweets... "))
def process_tweets(self, tweets_to_post):
    """Transforms tweets for posting them to Pleroma
    Expands shortened URLs
    Downloads tweet related media and prepares them for upload

    :param tweets_to_post: Dict of tweet objects to be processed
    :type tweets_to_post: dict
    :returns: Tweets ready to be published
    :rtype: list
    """
    # Remove RTs if include_rts is false
    if not self.include_rts:
        for tweet in tweets_to_post["data"][:]:
            try:
                for reference in tweet["referenced_tweets"]:
                    if reference["type"] == "retweeted":
                        tweets_to_post["data"].remove(tweet)
                        break
            except KeyError:
                pass
    # Remove replies if include_replies is false
    if not self.include_replies:
        for tweet in tweets_to_post["data"][:]:
            try:
                for reference in tweet["referenced_tweets"]:
                    if reference["type"] == "replied_to":
                        tweets_to_post["data"].remove(tweet)
                        break
            except KeyError:
                pass
    if self.hashtags:
        for tweet in tweets_to_post["data"][:]:
            try:
                tweet_hashtags = tweet["entities"]["hashtags"]
                i = 0
                while i < len(tweet_hashtags):
                    if tweet_hashtags[i]["tag"] in self.hashtags:
                        break
                    i += 1
                else:
                    tweets_to_post["data"].remove(tweet)
            except KeyError:
                tweets_to_post["data"].remove(tweet)
                pass

    for tweet in tweets_to_post["data"]:
        media = []
        tweet["text"] = _expand_urls(self, tweet)
        tweet["text"] = html.unescape(tweet["text"])
        if hasattr(self, "rich_text"):
            if self.rich_text:
                tweet["text"] = _replace_mentions(self, tweet)
        if self.nitter:
            tweet["text"] = _replace_nitter(self, tweet)

        try:
            for item in tweet["attachments"]["media_keys"]:
                for media_include in tweets_to_post["includes"]["media"]:
                    media_url = _get_media_url(
                        self, item, media_include, tweet
                    )
                    if media_url:
                        media.extend(media_url)
        except KeyError:
            pass
        # Create folder to store attachments related to the tweet ID
        tweet_path = os.path.join(self.tweets_temp_path, tweet["id"])
        os.makedirs(tweet_path, exist_ok=True)
        # Download media only if we plan to upload it later
        if self.media_upload:
            _download_media(self, media, tweet)
        # Process poll if exists and no media is used
        tweet["polls"] = _process_polls(self, tweet, media)

    return tweets_to_post


def _process_polls(self, tweet, media):
    try:
        if tweet["attachments"]["poll_ids"] and not media:

            # tweet_poll = tweet['includes']['polls']
            poll_url = f"{self.twitter_base_url_v2}/tweets"

            params = {
                "ids": tweet["id"],
                "expansions": "attachments.poll_ids",
                "poll.fields": "duration_minutes," "options",
            }

            response = requests.get(
                poll_url, headers=self.header_twitter, params=params
            )
            if not response.ok:
                response.raise_for_status()
            response_content = json.loads(response.content)
            tweet_poll = response_content["includes"]["polls"][0]

            pleroma_poll = {
                "options": [
                    option["label"] for option in tweet_poll["options"]
                ],
                "expires_in": tweet_poll["duration_minutes"] * 60,
            }

            # Add poll to tweet
            tweet["polls"] = pleroma_poll

        else:
            tweet["polls"] = None
    except KeyError:
        tweet["polls"] = None
        pass

    return tweet["polls"]


def _download_media(self, media, tweet):
    for idx, item in enumerate(media):
        if item["type"] != "video" and item["type"] != "animated_gif":
            media_url = item["url"]
        else:
            media_url = _get_best_bitrate_video(self, item)

        if media_url:
            response = requests.get(media_url, stream=True)
            try:
                if not response.ok:
                    response.raise_for_status()
            except requests.exceptions.HTTPError:
                if response.status_code == 404:
                    att_not_found = _(
                        "Exception occurred"
                        "\nMedia not found (404)"
                        "\n{tweet} - {media_url}"
                        "\nIgnoring attachment and continuing..."
                    ).format(tweet=tweet, media_url=media_url)
                    logger.warning(att_not_found)
                    return
                else:
                    response.raise_for_status()
            response.raw.decode_content = True
            filename = str(idx) + mimetypes.guess_extension(
                response.headers["Content-Type"]
            )
            file_path = os.path.join(
                self.tweets_temp_path, tweet["id"], filename
            )
            with open(file_path, "wb") as outfile:
                shutil.copyfileobj(response.raw, outfile)
            # Remove attachment if exceeds the limit
            if hasattr(self, "file_max_size"):
                file_size_bytes = os.stat(file_path).st_size
                max_file_size_bytes = parse_size(self.file_max_size)
                if file_size_bytes > max_file_size_bytes:
                    logger.error(
                        _(
                            "Attachment exceeded config file size limit ({})"
                        ).format(self.file_max_size)
                    )
                    logger.error(
                        _("File size: {}MB").format(
                            round(file_size_bytes / 2 ** 20, 2)
                        )
                    )
                    logger.error(_("Ignoring attachment and continuing..."))
                    os.remove(file_path)


def parse_size(size):
    units = {
        "B": 1,
        "KB": 2 ** 10,
        "MB": 2 ** 20,
        "GB": 2 ** 30,
        "TB": 2 ** 40,
    }
    size = size.upper()
    if not re.match(r" ", size):
        size = re.sub(r"([KMGT]?B)", r" \1", size)
    number, unit = [string.strip() for string in size.split()]
    return int(float(number) * units[unit])


def _replace_nitter(self, tweet):
    matching_pattern = "https://twitter.com"
    matches = re.findall(matching_pattern, tweet["text"])
    for match in matches:
        tweet["text"] = re.sub(match, self.nitter_base_url, tweet["text"])
    return tweet["text"]


def _replace_mentions(self, tweet):
    matches = re.findall(r"\B\@\w+", tweet["text"])
    for match in matches:
        mention_link = f"[{match}](https://twitter.com/{match[1:]})"
        tweet["text"] = re.sub(match, mention_link, tweet["text"])
    return tweet["text"]


def _expand_urls(self, tweet):
    # Replace shortened links
    try:
        for url_entity in tweet["entities"]["urls"]:
            matching_pattern = url_entity["url"]
            matches = re.findall(matching_pattern, tweet["text"])
            for match in matches:
                tweet["text"] = re.sub(
                    match, url_entity["expanded_url"], tweet["text"]
                )
    except KeyError:
        # URI regex
        matching_pattern = (
            r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]"
            r"{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*"
            r"\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{}"
            r';:\'".,<>?«»“”‘’]))'
        )
        matches = re.finditer(matching_pattern, tweet["text"])
        for matchNum, match in enumerate(matches, start=1):
            group = match.group()
            # don't be brave trying to unwound an URL when it gets
            # cut off
            if not group.__contains__("…"):
                if not group.startswith(("http://", "https://")):
                    group = f"http://{group}"
                session = requests.Session()  # so connections are
                # recycled
                response = session.head(group, allow_redirects=True)
                if not response.ok:
                    response.raise_for_status()
                expanded_url = response.url
                tweet["text"] = re.sub(
                    match.group(), expanded_url, tweet["text"]
                )
    return tweet["text"]


def _get_media_url(self, item, media_include, tweet):
    media_urls = []
    if item == media_include["media_key"]:
        # Video download not implemented in v2 yet
        # fallback to v1.1
        if (
            media_include["type"] == "video"
            or media_include["type"] == "animated_gif"
        ):
            tweet_video = self._get_tweets("v1.1", tweet["id"])
            xmd = tweet_video["extended_entities"]["media"]
            for extended_media in xmd:
                media_urls.append(extended_media)
            return media_urls
        else:
            media_urls.append(media_include)
            return media_urls


def _get_best_bitrate_video(self, item):
    bitrate = 0
    for variant in item["video_info"]["variants"]:
        try:
            if variant["bitrate"] >= bitrate:
                return variant["url"]
        except KeyError:
            pass
